/* doecm v1.10
Mikael Klasson
mklasson@acm.org
http://n137.ryd.student.liu.se/factors/
*/

#include <iostream>
#include <fstream>
#include <sstream>
#include <algorithm>
#include <string>
#include <cctype>
#include <ctime>
#include <vector>
#include <map>
#include <set>
using namespace std;

string version = "v1.10";

typedef unsigned int uint32;
typedef unsigned long long uint64;

map<string,string>	cfg;	//configuration data from .ini file and cmdline
map<string,bool>	okargs;	//allowed configuration arguments. <name,required>

struct factor {
	string	factorline,method,args;
	factor( string _factorline, string _method, string _args ) {
		factorline = _factorline; method = _method; args = _args;
	}
};

string stripws( string in ) {
	string	s;
	for( uint32 j = 0; j < in.size(); ++j ) {
		if( !isspace( in[j] ) ) {
			s = in.substr( j );
			break;
		}
	}
	while( s.size() && isspace( s[s.size() - 1] ) ) {
		s.resize( s.size() - 1 );
	}
	return s;
}

string tostring( long long n ) {
	stringstream ss;
	ss << n;
	return ss.str();
}

int toint( string s ) {
	stringstream ss( s );
	int n;
	ss >> n;
	return n;
}

uint64 touint64( string s ) {
	stringstream ss( s );
	uint64 n;
	ss >> n;
	return n;
}

bool isnumber( string s ) {
	for( int j = 0; j < s.size(); ++j ) {
		if( !isdigit( s[j] ) ) {
			return false;
		}
	}
	return true;
}

char tohex( int n ) {
	return string("0123456789abcdef")[n % 16];
}

string urlencode( string in ) {
	string	s;
	for( uint32 j = 0; j < in.size(); ++j ) {
		if( isalnum( in[j] ) ) {
			s += in[j];
		} else if( in[j] == ' ' ) {
			s += '+';
		} else {
			s += '%';
			s += tohex( in[j] / 16 );
			s += tohex( in[j] % 16 );
		}
	}
	return s;
}

// "11000000" -> "11e6"
string scientify( string n ) {
	int	e = 0;
	while( n.size() > 1 && n[n.size() - 1] == '0' ) {
		++e;
		n.resize( n.size() - 1 );
	}
	if( e < 4 ) {
		return n + string( e, '0' );
	}
	return n + "e" + tostring( e );
}

bool dump_factor( factor f ) {
	ofstream	fout( cfg["submitfailurefile"].c_str(), ios::app );
	fout << "#method(" << f.method << ")" << "args(" << f.args << ")" << endl;
	fout << f.factorline << endl;
	fout.close();
	return true;
}

bool submit_factor( string factorline, string method, string args, bool retryattempt ) {
	string	postdata = "name=" + urlencode( cfg["name"] ) 
		+ "&method=" + urlencode( method ) 
		+ "&factors=" + urlencode( factorline )
		+ "&args=" + urlencode( args )
		+ "&submitter=" + urlencode( "doecm " + version )
		;
	bool	failure = false;
	remove( cfg["wgetresultfile"].c_str() );
	cout << "Sending factor to server..." << flush;
	int r = system( ( cfg["wgetcmd"] + " -q --cache=off --output-document=\"" + cfg["wgetresultfile"] + "\" --post-data=\"" + postdata + "\" " + cfg["submiturl"] ).c_str() );
	cout << " Done." << endl;
	if( r != 0 ) {
		cout << "WARNING: wget returned " << r << ". There was probably an error." << endl;
		failure = true;
	}
	ifstream f( cfg["wgetresultfile"].c_str() );
	string	line;
	bool	found = false;
	while( getline( f, line ) ) {
		int pos = line.find( factorline );
		if( pos != line.npos ) {
			if( ( pos = line.find( "</td><td>", pos ) ) != line.npos ) {
				pos += string( "</td><td>" ).size();
				int last = line.find( "</td>", pos + 1 );
				if( last != line.npos ) {
					found = true;
					cout << "Result: " << line.substr( pos, last - pos ) << endl;
					break;
				}
			}
		}
	}
	f.close();
	if( !found ) {
		cout << "ERROR! Couldn't parse submission result." << endl;
		failure = true;
	}
	if( failure ) {
		if( !retryattempt ) {
			dump_factor( factor( factorline, method, args ) );
			cout << "Your factor has been saved in " << cfg["submitfailurefile"] << " for now." << endl;
		}
		return false;
	}
	return true;
}

bool report_work() {
	string	url = cfg["reporturl"] 
		+ "?method=" + urlencode( cfg["method"] )
		+ "&numbers=" + urlencode( cfg["numbers"] )
		+ "&nmin=" + urlencode( cfg["nmin"] )
		+ "&nmax=" + urlencode( cfg["nmax"] )
		+ "&b1=" + urlencode( cfg["b1"] )
		+ "&curves=" + urlencode( cfg["curves"] )
		;
	cout << "Reporting completed work..." << flush;
	int r = system( ( cfg["wgetcmd"] 
			+ " -q --cache=off --output-document=\"" 
			+ cfg["wgetresultfile"] 
			+ "\" \"" + url + "\"" ).c_str() );
	cout << " Done. Thanks!" << endl;
	if( r != 0 ) {
		cout << "WARNING: wget returned " << r << ". There was probably an error." << endl;
	}
	return r == 0;
}

bool download_composites() {
	string	url = cfg["compositeurl"] 
		+ "?enhanced=true&sort=" + urlencode( cfg["sort"] )
		+ "&nmin=" + urlencode( cfg["nmin"] )
		+ "&nmax=" + urlencode( cfg["nmax"] )
		+ "&order=" + urlencode( cfg["order"] )
		+ "&gzip=" + urlencode( cfg["use_gzip"] )
		+ "&recommendedwork=" + urlencode( cfg["recommendedwork"] )
		;
	stringstream	ss( cfg["numbers"] );
	int	k;
	char	plusminus,comma;
	int	filenr = 1;
	while( ss >> k >> plusminus ) {
		url += "&file" + tostring( filenr++ ) + "=" + tostring( k ) + "t2rn" + ( plusminus == '+' ? "p" : "m" ) + "1.txt";
		ss >> comma;
	}
	cout << "Downloading composites..." << flush;
	int r = system( ( cfg["wgetcmd"] 
			+ " -q --cache=off --output-document=\"" 
			+ cfg["compositefile"] 
			+ ( cfg["use_gzip"] == "yes" ? ".gz" : "" )
			+ "\" \"" + url + "\"" ).c_str() );
	cout << " Done." << endl;
	bool probablyerror = false;
	if( r != 0 ) {
		cout << "WARNING: wget returned " << r << ". There was probably an error." << endl;
		probablyerror = true;
	}
	if( cfg["use_gzip"] == "yes" ) {
		cout << "Unpacking composites..." << flush;
		r = system( ( cfg["gzipcmd"] + " -df " + cfg["compositefile"] + ".gz" ).c_str() );
		cout << " Done." << endl;
		if( r != 0 ) {
			cout << "WARNING: gzip returned " << r << ". There was probably an error." << endl;
			probablyerror = true;
		}
	}
	return probablyerror;
}

//returns true if all previously unsubmitted factors were submitted ok now
void process_unsubmitted_factors( bool forceattempt ) {
	static time_t	lastattempt = 0;
	if( !forceattempt && ( time( 0 ) - toint( cfg["submitretryinterval"] ) * 60 < lastattempt ) ) {
		return;
	}
	lastattempt = time( 0 );
	ifstream	fin( cfg["submitfailurefile"].c_str() );
	if( !fin.is_open() ) {
		//assume file doesn't exist
		return;
	}
	map<string,string>	info;
	info["method"] = cfg["method"];	//use this if method isn't stored in file
	string	line;
	vector<pair<factor,bool> >	unsubmitted;
	while( getline( fin, line ) ) {
		if( !line.size() ) {
			continue;
		}
		//is there extra info present?
		if( line[0] == '#' ) {
			stringstream	ss( line.substr( 1 ) );
			string	fn,val;
			while( getline( ss, fn, '(' ) ) {
				getline( ss, val, ')' );
				info[fn] = val;
			}
			continue;
		}
		unsubmitted.push_back( make_pair( factor( line, info["method"], info["args"] ), true ) );
	}
	fin.close();
	//no unsubmitted factors?
	if( !unsubmitted.size() ) {
		return;
	}
	cout << "Trying to send your unsubmitted factors..." << endl;
	int	succeeded = 0;
	for( vector<pair<factor,bool> >::iterator i = unsubmitted.begin(); i != unsubmitted.end(); ++i ) {
		if( submit_factor( i->first.factorline, i->first.method, i->first.args, true ) ) {
			++succeeded;
			i->second = false;
		}
	}
	string	failstring = "Please submit the factors in " + cfg["submitfailurefile"] + " manually\n"
			+ "or wait and see if doecm manages to submit them automatically later.\n"
			+ "Manual submit: " + cfg["manualsubmiturl"] + "\n";

	if( !succeeded ) {
		//none succeeded, just exit.
		cout << "Didn't manage to submit any factor now either." << endl;
		cout << failstring;
		return;
	}
	//remove all old failures
	remove( cfg["submitfailurefile"].c_str() );
	if( succeeded == unsubmitted.size() ) {
		cout << "Submitted all of your unsubmitted factors!" << endl;
		return;
	} else {
		cout << "Submitted " << tostring( succeeded ) << " of your " 
			<< unsubmitted.size() << " unsubmitted factors!" << endl;
		cout << failstring;
	}
	//dump remaining unsubmitted to file
	for( vector<pair<factor,bool> >::iterator i = unsubmitted.begin(); i != unsubmitted.end(); ++i ) {
		if( i->second ) {
			dump_factor( i->first );
		}
	}

	return;
}

bool cfg_set( string src, string arg, string val ) {
	if( !okargs.count( arg ) ) {
		cout << "WARNING: unrecognized option: '" << arg << "'" << endl;
		return false;
	}
	cout << src << ": " << arg << " = " << val << endl;
	cfg[arg] = val;
	return true;
}

bool parse_cmdline( int argc, char ** argv ) {
	for( int j = 1; j < argc; ) {
		if( argv[j][0] == '-' ) {
			string	arg = string( argv[j] ).substr( 1 );
			if( arg == "h" || arg == "-help" || arg == "?" ) {
				return false;	// just display help info, then exit
			}
			if( j + 1 >= argc ) {
				cout << "ERROR: cmdline switch without argument: " << argv[j] << endl;
				return false;
			}
			cfg_set( "cmdline", arg, argv[j + 1] );
			j += 2;
		} else {
			cout << "WARNING: unrecognized cmdline argument: " << argv[j] << endl;
			++j;
		}
	}
	return true;
}

bool read_inifile( string fname ) {
	ifstream f( fname.c_str() );
	string	line,arg,val;

	if( !f.is_open() ) {
		cout << "ERROR: couldn't open .ini file: " << fname << endl;
		cout << "Look in the doecm archive for an example of a proper .ini file." << endl;
		return false;
	}
	while( getline( f, line ) ) {
		stringstream	ss( line );
		getline( ss, arg, '=' );
		ss >> ws;
		arg = stripws( arg );
		if( !getline( ss, val ) ) {
			continue;
		}
		val = stripws( val );
		if( arg.substr( 0, 2 ) == "//" || arg.substr( 0, 1 ) == "#" || arg.substr( 0, 1 ) == ";" ) {
			continue;
		}
		cfg_set( "ini", arg, val );
	}
	f.close();
	return true;
}

bool init_args() {
	string reqargs[] = {	"name", "b1", "b1increase", "curves", "nmin", "nmax", "loop", "numbers", 
				"autosubmit", "autodownload", "reportwork", "wgetcmd", "ecmcmd", "gzipcmd", 
				"use_gzip", "sort", "order", "compositefile", "compositeurl", "submiturl", 
				"manualsubmiturl", "reporturl", "factorfile", "submitfailurefile", "sigmafile", 
				"wgetresultfile", "ecmresultfile", "recommendedwork", "method", "submitretryinterval"
	};
	string optargs[] = { "ecmargs", "fallback" };
	for( int j = 0; j < sizeof( reqargs ) / sizeof( string ); ++j ) {
		okargs[reqargs[j]] = true;
	}
	for( int j = 0; j < sizeof( optargs ) / sizeof( string ); ++j ) {
		okargs[optargs[j]] = false;
	}
	return true;
}

bool verify_args() {
	bool	ok = true;
	for( map<string,bool>::iterator i = okargs.begin(); i != okargs.end(); ++i ) {
		if( i->second && cfg[i->first] == "" ) {
			cout << "ERROR: missing required setting: " << i->first << endl;
			ok = false;
		}
	}
	if( toint( cfg["nmin"] ) > toint( cfg["nmax"] ) ) {
		cout << "ERROR: nmin=" << cfg["nmin"] << " > nmax=" << cfg["nmax"] << " doesn't make sense." << endl;
		ok = false;
	}
	if( cfg["method"] == "P-1" ) {
		cfg["ecmargs"] += " -pm1";
		if( toint( cfg["curves"] ) != 1 ) {
			cout << "Running " << cfg["curves"] << " P-1 tests doesn't make sense. Forcing <curves> to 1." << endl;
			cfg["curves"] = 1;
		}
	} else if( cfg["method"] == "P+1" ) {
		cfg["ecmargs"] += " -pp1";
		if( toint( cfg["curves"] ) > 3 ) {
			cout << "WARNING: running " << cfg["curves"] << " P+1 tests is probably a waste. Try 3 instead." << endl;
		}
	}
	return ok;
}

void found_factor( string p, bool enhanced, string expr, string inputnumber, uint64 sigma ) {
	ofstream	fout( cfg["factorfile"].c_str(), ios::app );
	stringstream	factorline;
	if( enhanced ) {
		factorline << p << " | " << expr;
	} else {
		factorline << p << " | " << inputnumber;
	}
	cout << factorline.str() << "\t" << "(" << tostring( p.size() ) << " digits)" << endl;
	fout << factorline.str() << endl;
	if( cfg["autosubmit"] == "yes" ) {
		string	args = "B1=" + scientify( cfg["b1"] );
		if( cfg["method"] == "ECM" ) {
			args += ",s=" + tostring( sigma );
		}
		if( submit_factor( factorline.str(), cfg["method"], args, false ) ) {
			//it worked! try to send any unsubmitted factors while we're at it
			process_unsubmitted_factors( true );
		}
	}
	fout.close();
}

// returns the number of composites in <compositefile>
// also shuffles them if "order = random"
// processes recommended work settings from server if "recommendedwork = yes"
int init_composites() {
	int cnt = 0;
	string	line;
	cout << "Initializing..." << endl;
	if( cfg["recommendedwork"] == "yes" ) {
		ifstream f( cfg["compositefile"].c_str() );
		getline( f, line );
		if( line.substr( 0, 1 ) != "#" ) {
			cout << "WARNING: trying to do recommended work, but didn't get any settings from server." << endl;
		} else {
			stringstream	ss( line.substr( 1 ) );
			ss >> ws;
			string	fn,val;
			while( getline( ss, fn, '(' ) ) {
				getline( ss, val, ')' );
				cfg_set( "recommended work", fn, val );
			}
		}
		f.close();
	}
	ifstream f( cfg["compositefile"].c_str() );
	if( cfg["order"] == "random" ) {
		vector<pair<string,string> >	v;
		string	comment;
		while( getline( f, line ) ) {
			if( line.substr( 0, 1 ) != "#" ) {
				++cnt;
				v.push_back( make_pair( line, comment ) );
				comment = "";
			} else {
				comment = line;
			}
		}
		f.close();
		random_shuffle( v.begin(), v.end() );
		ofstream	fout( cfg["compositefile"].c_str() );
		for( uint32 j = 0; j < v.size(); ++j ) {
			if( v[j].second != "" ) {
				fout << v[j].second << endl;
			}
			fout << v[j].first << endl;
		}
		fout.close();
	} else {
		while( getline( f, line ) ) {
			if( line.substr( 0, 1 ) != "#" ) {
				++cnt;
			}
		}
		f.close();
	}
	return cnt;
}

int main( int argc, char ** argv ) {
	cout << "doecm " << version << " by Mikael Klasson (mklasson@acm.org)" << endl;
	cout << "usage: doecm [-b1 <B1> -curves <curves> ...]" << endl;
	cout << "  Make sure you're using the right username." << endl;
	cout << "  Look in doecm.ini for available options." << endl;
	srand( time( 0 ) );
	init_args();
	read_inifile( "doecm.ini" );
	if( !parse_cmdline( argc, argv ) ) {
		return 0;
	}
	if( !verify_args() ) {
		cout << "Fix your settings and try again. Exiting." << endl;
		return 1;
	}
	cout << "Using factorization method " << cfg["method"] << endl;

	int	totalfactors = 0;
	do {
		process_unsubmitted_factors( true );

		if( cfg["autodownload"] == "yes" ) {
			download_composites();
		}

		int ccnt = init_composites();
		cout << "Found " << ccnt << " composites in " << cfg["compositefile"] << "." << endl;
		if( !ccnt ) {
			//compositefile is empty
			if( cfg["fallback"] == "yes" && cfg["recommendedwork"] != "yes" ) {
				cout << "Switching to fallback mode." << endl;
				cfg["recommendedwork"] = "yes";
				cfg["autodownload"] = "yes";
				continue;
			} else {
				cout << "Work complete. Exiting." << endl;
				break;
			}
		}
		ifstream fin( cfg["compositefile"].c_str() );
		bool	enhanced = false;
		string	line,expr;
		while( getline( fin, line ) ) {
			if( line.substr( 0, 1 ) == "#" ) {
				stringstream	ss( line.substr( 1 ) );
				ss >> ws;
				string	fn,val;
				while( getline( ss, fn, '(' ) ) {
					getline( ss, val, ')' );
					if( fn == "expr" ) {
						expr = val;
						enhanced = true;
					}
				}
				continue;
			}
			string	inputnumber = line;
			if( enhanced ) {
				cout << "Factoring " << expr << " c" << inputnumber.size() << endl;
			} else {
				cout << "Factoring " << inputnumber << endl;
			}
			system( ( "echo " + inputnumber + " | " + cfg["ecmcmd"] + " -c " + cfg["curves"] + " " + cfg["ecmargs"] + " " + cfg["b1"] + " > " + cfg["ecmresultfile"] ).c_str() );
			ifstream ftmp( cfg["ecmresultfile"].c_str() );
			uint64 sigma = 0;
			while( getline( ftmp, line ) ) {
				int pos = line.find( "sigma=" );
				if( pos != line.npos ) {
					sigma = touint64( line.substr( pos + 6 ) );
				}
				pos = line.find( ": " );
				if( pos != line.npos ) {
					++totalfactors;
					string p = line.substr( pos + 2 );
					ofstream flog( cfg["sigmafile"].c_str(), ios::app );
					flog << p << " sigma=" << sigma << endl;
					flog.close();

					found_factor( p, enhanced, expr, inputnumber, sigma );
					bool trialfactor = true;
					if( trialfactor && p.size() <= 10 ) {
						uint64	n = touint64( p );
						for( uint64 f = 3; f * f <= n; f += 2 ) {
							if( !( n % f ) ) {
								found_factor( tostring( f ), enhanced, expr, inputnumber, sigma );
								do {
									n /= f;
								} while( !( n % f ) );
							}
						}
						if( n > 1 && n != touint64( p ) ) {
							found_factor( tostring( n ), enhanced, expr, inputnumber, sigma );
						}
					}

					cout << "#factors found: " << totalfactors << endl;
					break;
				}
			}
			ftmp.close();
			process_unsubmitted_factors( false );
		}
		fin.close();
		cout << "#factors found: " << totalfactors << endl;
		if( cfg["reportwork"] == "yes" ) {
			report_work();
		}
		//increase b1
		cfg["b1"] = tostring( touint64( cfg["b1"] ) + touint64( cfg["b1increase"] ) );
		cout << "Increasing B1 to " << cfg["b1"] << endl;
	} while( cfg["loop"] == "yes" );

	//do one last valiant attempt to submit any unsubmitted factors
	process_unsubmitted_factors( true );

	remove( cfg["wgetresultfile"].c_str() );
	remove( cfg["ecmresultfile"].c_str() );
}
